/*
 * leds-input.c - Make LEDs controlled by the input layer appear as devices in
 *                the LED class.
 *
 * Copyright (C) 2010 Adam Nielsen <a.nielsen@shikadi.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301 USA.
 *
 * Known issues:
 *
 *   - Due to the limited bandwidth available to most input devices, it's
 *     possible for a fast enough trigger (e.g. netfilter) to saturate the
 *     link with LED on/off messages, causing a delay between pressing a key
 *     and having it appear on the screen.  With netfilter the "--led-delay 1"
 *     option solves the problem, but it might be worth implementing a delay
 *     here too so that it works with other LED trigger sources too.
 *
 *   - The input layer itself doesn't share the LEDs nicely.  If you use a LED
 *     for a slow notification (e.g. netfilter with --led-delay 1000 attached
 *     to the scroll lock LED), pressing an unrelated key like caps lock will
 *     also turn the scroll lock off.  Hopefully it's just a minor annoyance
 *     though!
 */

#include <linux/module.h>
#include <linux/input.h>
#include <linux/leds.h>
#include <linux/err.h>

#define DRVNAME	"leds-input"
#define DRVMSG	DRVNAME ": "

struct handler_data {
	struct input_handle handle;
	struct list_head led_list;
};

struct leddev {
	struct list_head list;
	struct led_classdev led_cdev;
	struct handler_data *handler_data;
	int evdev_led_id;
	char led_name[20]; /* As per led_classdev.name */
};

/* The order of these mappings must match the LED_* constants in input.h */
static const char *const lednames[] = {
	"num",      /* LED_NUML */
	"caps",     /* LED_CAPSL */
	"scroll",   /* LED_SCROLL */
	"kana",     /* LED_KANA */
	"compose",  /* LED_COMPOSE */
	"sleep",    /* LED_SLEEP */
	"suspend",  /* LED_SUSPEND */
	"mute",     /* LED_MUTE */
	"misc",     /* LED_MISC */
	"mail",     /* LED_MAIL */
	"charging", /* LED_CHARGING */
	"user12",
	"user13",
	"user14",
	"user15",
	"user16",   /* LED_MAX */
};


/* Called when the LED class driver wants this LED to change brightness */
static void leds_input_set(struct led_classdev *led_cdev,
	enum led_brightness value)
{
	const struct leddev *leddev;

	leddev = container_of(led_cdev, struct leddev, led_cdev);
	input_inject_event(&leddev->handler_data->handle,
		EV_LED,
		leddev->evdev_led_id,
		(value > LED_OFF) ? 1 : 0
	);
	return;
}

/* Called when the LED class driver wants to know this LED's brightness */
static enum led_brightness leds_input_get(struct led_classdev *led_cdev)
{
	const struct leddev *leddev =
		container_of(led_cdev, struct leddev, led_cdev);
	struct input_dev *indev;

	if (!leddev->handler_data)
		return LED_OFF; /* too early */
	indev = leddev->handler_data->handle.dev;
	if (!indev)
		return LED_OFF;

	if (test_bit(leddev->evdev_led_id, indev->led))
		return LED_FULL;

	return LED_OFF;
}

/* Called once per LED on each input-layer device which has LEDs */
static int leddev_create(struct handler_data *handler_data,
	struct input_handler *handler, struct input_dev *dev, int evdev_led_id
) {
	int error, max_prefix_len;
	struct leddev *leddev;

	leddev = kzalloc(sizeof(struct leddev), GFP_KERNEL);
	if (!leddev) {
		error = -ENOMEM;
		goto err;
	}

	/* A LED is called something like "device:colour:name" such as
	 * "keyboard:green:capslock".  Since it's limited to 20 chars and we
	 * don't know what colour it might be, we leave the colour blank ("::")
	 * and put the LED name (caps, scroll, etc.) as per the input layer
	 * constant, and leave whatever's left of the 20 chars for the device
	 * name/id at the start.
	 */
	max_prefix_len = sizeof(leddev->led_name) - strlen(lednames[evdev_led_id]) - 3;
	snprintf(leddev->led_name, max_prefix_len, "in%04X", dev->id.product);
	strcat(leddev->led_name, "::");
	strcat(leddev->led_name, lednames[evdev_led_id]);

	leddev->led_cdev.name = leddev->led_name;
	leddev->led_cdev.brightness_set = leds_input_set;
	leddev->led_cdev.brightness_get = leds_input_get;
	error = led_classdev_register(&dev->dev, &leddev->led_cdev);
	if (error)
		goto err_free_mem;

	leddev->handler_data = handler_data;
	leddev->evdev_led_id = evdev_led_id;
	INIT_LIST_HEAD(&leddev->list);

	/* Add the new LED onto the input_handler's list of LEDs for this input device */
	list_add(&leddev->list, &handler_data->led_list);

	return 0;

err_free_mem:
printk(KERN_ERR DRVMSG "error %d\n", error);

	kfree(leddev);
err:
	printk(KERN_ERR DRVMSG "Unable to register LED device\n");
	return error;
}

/* This function is called when a device has been found by the input layer
 * with LEDs on it.
 */
static int leddev_connect(struct input_handler *handler, struct input_dev *dev,
	const struct input_device_id *id
) {
	int error, i, newleds;
	struct handler_data *handler_data;

	printk(KERN_DEBUG DRVMSG "Input connect for device %s (phys %s, uniq %s)\n", dev->name, dev->phys, dev->uniq);

	handler_data = kzalloc(sizeof(struct handler_data), GFP_KERNEL);
	if (!handler_data) {
		error = -ENOMEM;
		goto err;
	}

	handler_data->handle.dev = input_get_device(dev);
	handler_data->handle.name = "leddev";
	handler_data->handle.handler = handler;
	handler_data->handle.private = handler_data;
	INIT_LIST_HEAD(&handler_data->led_list);

	error = input_register_handle(&handler_data->handle);
	if (error)
		goto err_free_mem;

	error = input_open_device(&handler_data->handle);
	if (error)
		goto err_unregister;

	/* Register a device in the LED class for each LED reported by the input layer */
	newleds = 0;
	for (i = 0; i < LED_CNT; i++) {
		if (test_bit(i, dev->ledbit)) {
			//printk(KERN_DEBUG DRVMSG "has led %d (%s)\n", i, lednames[i]);
			error = leddev_create(handler_data, handler, dev, i);
			if (!error)
				newleds++;
		}
	}

	/* Only return an error if no LEDs at all could be created */
	if (newleds == 0)
		goto err_close_dev;

	return 0;
err_close_dev:
	input_close_device(&handler_data->handle);
err_unregister:
	input_unregister_handle(&handler_data->handle);
	input_put_device(dev);
err_free_mem:
	kfree(handler_data);
err:
	printk(KERN_ERR DRVMSG
		"No LED devices could be created for device %s\n", dev->name);
	return error;
}

static void leddev_disconnect(struct input_handle *handle)
{
	struct handler_data *handler_data = handle->private;
	struct leddev *leddev, *leddev_tmp;

	printk(KERN_DEBUG DRVMSG "Removing LED devices from input device\n");
	list_for_each_entry_safe(leddev, leddev_tmp, &handler_data->led_list, list) {
		//printk(KERN_DEBUG DRVMSG "Removing LED device %s\n", leddev->led_name);
		led_classdev_unregister(&leddev->led_cdev);
		kfree(leddev);
		/* Don't bother removing the deleted item from the list,
		 * because when we're done the list won't exist any more.
		 */
	}
	input_close_device(&handler_data->handle);
	input_unregister_handle(&handler_data->handle);
	input_put_device(handler_data->handle.dev);
	kfree(handler_data);
}

static void leddev_event(struct input_handle *handle,
	unsigned int type, unsigned int code, int value)
{
	/* We're not interested in receiving events about LEDs changing here, but
	 * input_handler.event can't be set to NULL!
	 */
	/*if (type == EV_LED) {
		if (code < LED_CNT) {
			printk(KERN_DEBUG DRVMSG "code %d val %d name %s\n",
				code, value, lednames[code]);
		} else {
			printk(KERN_DEBUG DRVMSG "code %d val %d - code > LED_CNT!\n",
				code, value);
		}
	}*/
	return;
}

static const struct input_device_id leddev_ids[] = {
	{
		.flags = INPUT_DEVICE_ID_MATCH_EVBIT,
		.evbit = { BIT_MASK(EV_LED) },
	},   /* Any device with one or more LEDs */
	{ }  /* End of list */
};

static struct input_handler leddev_handler = {
	.event = leddev_event,
	.connect = leddev_connect,
	.disconnect = leddev_disconnect,
	.name = "leddev",
	.id_table = leddev_ids,
};

static int __init leds_input_init(void)
{
	return input_register_handler(&leddev_handler);
}

static void __exit leds_input_exit(void)
{
	input_unregister_handler(&leddev_handler);
}

module_init(leds_input_init);
module_exit(leds_input_exit);

MODULE_AUTHOR("Adam Nielsen <a.nielsen@shikadi.net>");
MODULE_DESCRIPTION("Make LEDs on input-layer devices appear as individual "
	"devices in the LED class");
MODULE_LICENSE("GPL");
